package iSATLibrary;

import java.util.ArrayList;
import java.util.BitSet;

import com.sun.jna.Pointer;
import com.sun.jna.ptr.IntByReference;

/**
 * This class is for using incremental SAT library API (IPASIR) and its extended API available from Java.
 * This wraps these APIs. Use the methods of this class when using incremental SAT APIs.
 * @author Tatsuya Sako
 */
public class ISatSolver extends ISatLibrary {

    private static ArrayList<int[]> buffer = new ArrayList<int[]>(); // temporary storage for clauses to add
    private static int countOfClauses = 0;    
    private int limit; // upper limit number of clauses to add SAT solver
	
    private String solverName;
        
	public ISatSolver(String name){
		super(name);
		this.solverName = name;
		this.limit = 1000;
	}
	
	
	/**
	 * Gets a signature of SAT solver.
	 * @return the signature of SAT solver
	 */
	public String ipasir_signature(){
		return ipasirLib.ipasir_signature();
	}
	
	/**
	 * Constructs a new SAT solver.
	 * @param name the name of SAT solver
	 * @return solver's object
	 */
	public ISatSolver ipasir_init(String name){
		ISatSolver solver = new ISatSolver(name);
		return solver;
	}
	
	/**
	 * Releases the solver.
	 */
	public void ipasir_release(){
		ipasirLib.ipasir_release(solverPtr);
	}
	
	/**
	 * Adds a literal to SAT solver or finarizes the clause with a "0".
	 * @param lit_or_zero a literal or "0"
	 */
	public void ipasir_add(int lit_or_zero){
		ipasirLib.ipasir_add(solverPtr, lit_or_zero);
	}
	
	/**
	 * Adds an assumption.
	 * @param lit the assumption
	 */
	public void ipasir_assume(int lit){
		ipasirLib.ipasir_assume(solverPtr, lit);
	}

	 /**
	  * Solves a SAT problem.
	  * @return true (if the problem is SAT), false (UNSAT).
	  */
	 public int ipasir_solve(){
		 return ipasirLib.ipasir_solve(solverPtr);
	 }	
	
	/**
	 * Gets the truth value of the given literal in the assignment.
	 * @param lit the literal 
	 * @return the truth value of the given literal
	 */
	public int ipasir_val(int lit){
		return ipasirLib.ipasir_val(solverPtr, lit);
	}

	/**
	 * Checks if the given assumption was used to prove UNSAT in the last search.
	 * @param lit the assumption literal
	 * @return 1 (used), 0 (otherwise)
	 */
	public int ipasir_failed(int lit){
		return ipasirLib.ipasir_failed(solverPtr, lit);
	}

	/**
	 * Sets a number of clauses added to SAT solver at a time. 
	 * Default value : 1000
	 * @param limit the number of clauses added to SAT solver at a time 
	 */
	public void set_Limit(int limit){
		this.limit = limit;
	}
	
	/**
	 * Makes an array for buffering clauses.
	 * This method appends a new clause to the end of the buffer. 
	 * @return appended buffer
	 */
	public int[] makeBuffer(){
		int length = 0;
		int offset = 0;
		for(int[] array : buffer){
			length += (array.length+1); // for adding "0"
		}
		int[] margedBuffer = new int[length];
		for(int[] array : buffer){
			int l = array.length;
			for(int i=0; i<l; i++){
				margedBuffer[offset+i] = array[i]; 
			}
			margedBuffer[offset+l] = 0;
			offset = offset+l+1;
		}
		buffer.clear();
		return margedBuffer;
	}
	
	/**
	 * Sets the verbosity of SAT solver's output. 
	 * If you set "0", solver's output isn't displayed.
	 * @param verbosity the verbosity of SAT solver's output
	 */
	public void set_verbosity(int verbosity){
		isatLib.isat_set_verbosity(solverPtr, verbosity);
	}
	
	/**
	 * Gets the number of variables in SAT solver.
	 * @return number of variables
	 */
	public int get_vars(){
		return isatLib.isat_vars(solverPtr);
	}
	
	/**
	 * Adds a clause to SAT solver.
	 * You must add "0" to the end of the array "clause". 
	 * @param clause clause
	 */
	public void add_a_clause(int[] clause){
		isatLib.isat_add_clauses(solverPtr, clause, clause.length);
	}
	
	/**
	 * Adds multiple clauses to SAT solver at a time.
	 * Clauses are temporally buffered.
	 * @param lits array of literals
	 */
	public void addNormalizedClause(int[] lits) {
		buffer.add(lits.clone()); 
		countOfClauses++;
		if(countOfClauses==limit){
			int[] clauses = makeBuffer();
			add_a_clause(clauses);
			countOfClauses=0;
        }
	}
	
	/**
	 * Adds a clause to SAT solver.
	 * @param clause clause
	 */
	 public void add_clause(int[] clause){
		 addNormalizedClause(clause);
	 }

	 /**
	  * Adds rest clauses in buffer to SAT solver.
	  * This method is called before solving a SAT problem.
	  */
	 public void done(){
		 if(countOfClauses == 0) return;
		 int[] clauses = makeBuffer();
		 add_a_clause(clauses);
		 countOfClauses = 0;
	 }
	 
	 /**
	  * Solves a SAT problem.
	  * @return true (if the problem is SAT), false (UNSAT).
	  */
	 public boolean solve(){
		 done();
		 return (ipasirLib.ipasir_solve(solverPtr)==10) ? true : false;
	 }	

	 /**
	  * Solves a SAT problem with assumption.
	  * @param assumption assumption 
	  * @return true (if the problem is satisfiable(SAT) ), false (otherwise).
	  */	
	 public boolean solve(int[] assumption){
		 for (int i : assumption)
			 ipasirLib.ipasir_assume(solverPtr,i);
		 return solve();
	 }	
	 
	 /**
	  * Adds a blocking clause to SAT solver.
	  * @param clause the clause required to block
	  */
	 public void add_blocking_clause(int[] clause){
		 isatLib.isat_add_blocking_clause(solverPtr, clause, clause.length);
	 }
	
	/**
	 * Adds a blocking clause to SAT solver.
	 * This blocking clause contains all SAT variables and blocks obtained solution only.
	 */
	public void add_blocking_clause(){
		isatLib.isat_add_blocking_clause(solverPtr, null, 0);
	}

	/**
	 * Excludes a literal from the simplification of SAT solver.
	 * This method is available only when SAT solver uses simplification process.
	 * If you use the SAT solver which does "simplification", you should freeze the literal.
	 * @param literal the literal required to freeze
	 */
	public void set_frozen(int literal){
		isatLib.isat_freeze(solverPtr, literal);
	}
	
	/**
	 * Checks if the literal is frozen.
 	 * This method is available only when SAT solver uses simplification process.
	 * @param literal the literal required to check if frozen
	 * @return true (if the literal is frozen), false (otherwise)
	 */
	public boolean is_eliminated(int literal){
		return isatLib.isat_is_eliminated(solverPtr, literal);
	}
	
	/**
	 * Gets SAT solution (assignments to SAT variables) by bitset.
	 * If you need to decode to CSP solution by Sugar, you should use this method.
	 * @return the model of solution by bitset
	 */
	public BitSet get_model() {
	    BitSet model = new BitSet();
		IntByReference modelptr = new IntByReference();
		modelptr = isatLib.isat_model(solverPtr);
		Pointer p = modelptr.getPointer();
		for(int i=0,v = isatLib.isat_vars(solverPtr); i<v; i++){
			if(p.getInt(i*4) > 0)
				model.set(p.getInt(i*4));
		}
		isatLib.isat_clean(solverPtr, p);
        return model;
	}
	
	/**
	 * Gets SAT solution (assignments to SAT variables) by binary integer array.
	 * @return the model of solution by binary integer array (true:1, false:0)
	 */
	public int[] model(){
		int length = (int)isatLib.isat_vars(solverPtr);
		int[] model = new int[length];
		IntByReference modelptr = new IntByReference();
		modelptr = isatLib.isat_model(solverPtr);
		Pointer p = modelptr.getPointer();
		for(int i=0; i<length; i++)
			model[i] = p.getInt(i*4)>0 ? 1 : 0;
		isatLib.isat_clean(solverPtr, p);
		return model;
	}

	/**
	 * Gets SAT solution (assignments to SAT variables) by integer array.
	 * @return the model of solution by integer array
	 */
	public int[] get_int_model(){
		int length = (int)isatLib.isat_vars(solverPtr);
		int[] model = new int[length];
		IntByReference modelptr = new IntByReference();
		modelptr = isatLib.isat_model(solverPtr);
		Pointer p = modelptr.getPointer();
		for(int i=0; i<length; i++)
			model[i] = p.getInt(i*4);
		isatLib.isat_clean(solverPtr, p);
		return model;
	}
	
	/**
	 * Outputs SAT solution on standard output.
	 */
	public void output_model(){
		BitSet model = get_model();
		for(int i=1; i<=isatLib.isat_vars(solverPtr); i++)
			System.out.print(model.get(i) ? i+" " : -i+" ");
		System.out.println();
	}

	/*
	 * Gets time that SAT solver used on this solving.
	 * @return solving time
	 */
	/*public double get_solve_time() {
		 return isatLib.isat_get_time(solverPtr);
	}*/

	/**
	 * Gets total time of invocation of SAT solver.
	 * @return elapsed time
	 */
	public double get_time() {
		 return isatLib.isat_get_time(solverPtr);
	}

	/**
	 * Gets a name of SAT solver.
	 * @return the name of SAT solver
	 */
	public String get_solver_name() {
		 return solverName;
	}

}
